/* A generic server module. Basically handles setting up a select loop
   and manages file descriptors. */

#if defined(ultrix) && defined(UPDATEPASSWD)
/* 4096? Come on. Copying fd_set structures suddenly becomes reasonable
   with only 64 descriptors defined. */
#define FD_SETSIZE 128
#endif
#include <stdio.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#if defined(UPDATEPASSWD) || defined(USE_BP) || defined(ISKERBEROS)
#include <krb.h>
#include "update_passwd.h"
#else
#include "mymalloc.h"
typedef int (*PFI)();
#endif
#include <syslog.h>
#ifdef _AIX
#include <sys/select.h>
#endif
#include "fdlist.h"
#define invaliddes(d) ((d)<minfds || (d)>maxfds || fdlist[d]==NULL)

static fd_set read_mask,write_mask,except_mask;
static struct fdlist **fdlist=NULL;
static int maxfds=0,minfds=0-1;
extern int debug;

/* Initialize things. */

init_server() {
  FD_ZERO(&read_mask);
  FD_ZERO(&write_mask);
  FD_ZERO(&except_mask);
}

add_link_to_des(ptr)
     struct fdlist *ptr;
{
  ptr->linkcnt++;
}

remove_link_from_des(ptr)
     struct fdlist *ptr;
{
  ptr->linkcnt--;
  if(ptr->linkcnt==0) free(ptr);
}

/* Add a descriptor to be polled. D is the descriptor, events are the
   events it is interested in (1=read, 2=write, 4=accept), fun is
   the function that should normally be called when an event is received,
   addr is an optional pointer to the address of the socket (it will
   be kept in the fdlist pointer), and fd is an optional FILE *
   descriptor that is associated with descriptor d. */

add_des(d,events,fun,addr,fd)
     int d,events;
     PFI fun;
     struct sockaddr_in *addr;
     FILE *fd;
{
  if((d+1)>maxfds) {
    int x;

    x=maxfds;
    maxfds=d+1;
    if(fdlist==NULL) fdlist=tnmalloc(struct fdlist *,maxfds);
    else tnrealloc(fdlist,struct fdlist *,maxfds);
    for(;x<maxfds;x++) fdlist[x]=NULL;
  }
  if(minfds==(0-1)) minfds=d;
  if(d<minfds) minfds=d;
  fdlist[d]=tmalloc(struct fdlist);
  fdlist[d]->linkcnt=1;
  if(debug) fprintf(stderr,"Adding des %d minfds=%d\n",d,minfds);
  fdlist[d]->events=events;
  set_mask_to_events(d,events);
  fdlist[d]->defaultfun=fun;
  fdlist[d]->funlist=NULL;
  fdlist[d]->state=NULL;
  fdlist[d]->closefun=NULL;
  fdlist[d]->state=0;
  fdlist[d]->lockcnt=0;
  fdlist[d]->fd=fd;
  fdlist[d]->userdata=NULL;
  fdlist[d]->des=d;
  if(addr!=NULL) fdlist[d]->sock_addr=(*addr);
  return(0);
}

lockdes(d)
     int d;
{
  if(invaliddes(d)) {
    syslog(LOG_ERR,"lockdes(%d) called on nonvalid descriptor",d);
    return(1);
  }
  else {
    if(fdlist[d]->lockcnt++>0) return;
    clear_mask_from_des(d);
    return(0);
  }
}

unlockdes(d)
     int d;
{
  if(invaliddes(d)) {
    syslog(LOG_ERR,"unlockdes(%d) called on nonvalid descriptor",d);
    return(1);
  }
  else {
    if(fdlist[d]->lockcnt<=0) {
      syslog(LOG_ERR,"unlockdes(%d) called on unlocked des",d);
    }
    else {
      fdlist[d]->lockcnt--;
      if(fdlist[d]->lockcnt==0) {
	set_mask_to_events(d,fdlist[d]->currevents);
      }
    }
  }
  return(0);
}

/* This function changes the default function for descriptor d. */
set_fun(d,fun)
     int d;
     PFI fun;
{
  if(invaliddes(d)) return(1);
  fdlist[d]->defaultfun=fun;
  return(0);
}

/* set_close_fun() sets a function to be called when descriptor d is closed. */

set_close_fun(d,fun)
     int d;
     PFI fun;
{
  if(invaliddes(d)) return(1);
  fdlist[d]->closefun=fun;
  return(0);
}

static set_mask_to_events(d,events)
     int d,events;
{
  if(invaliddes(d)) return(1);
  clear_mask_from_des(d);
  if(events==0) {
    syslog(LOG_ERR,"Null events in set_mask_to_events()");
  }
  if(events&1) FD_SET(d,&read_mask);
  if(events&2) FD_SET(d,&write_mask);
  FD_SET(d,&except_mask); /* We want the select loop to always be notified
			     of "exceptional conditions" */
  fdlist[d]->currevents=events;
  return(0);
}

static clear_mask_from_des(d)
     int d;
{
  if(invaliddes(d)) return(1);
  FD_CLR(d,&read_mask);
  FD_CLR(d,&write_mask);
  FD_CLR(d,&except_mask);
  return(0);
}

/* This function adds a function to be called whenever events are received
   on descriptor d. The function should delete itself with
   remove_fun_from_des() when it is done. If the function returns a non-zero
   value, it is assumed that an error has occured and the descriptor is
   closed. This function is called in preference to the default one, and only
   the first function in the list is called. */

add_fun_to_des(d,fun,events)
     int d;
     PFI fun;
     int events;
{
  struct funlist **p,*prev;

  if(invaliddes(d)) return(1);
  p=(&(fdlist[d]->funlist));
  if((*p)==NULL && events) {
    if(debug) fprintf(stderr,"Adding first fun to %d\n",d);
    set_mask_to_events(d,events);
  }
  while((*p)!=NULL) p=(&((*p)->next));
  (*p)=tmalloc(struct funlist);
  (*p)->next=NULL;
  (*p)->fun=fun;
  (*p)->events=events;
  return(0);
}

/* This function removes the first function in the function list for
   descriptor d. It is normally called from within the event function itself
   when it is finished with an abrt value of 0. abrt=1 indicates it is being
   called due to a failure or a close on the descriptor. */

remove_fun_from_des(d,abrt)
     int d;
     int abrt;
{
  struct funlist *l;

  if(invaliddes(d)) return(1);
  if(fdlist[d]->funlist==NULL) return;
  if(abrt) fdlist[d]->funlist->fun(d,fdlist[d],1);
  l=fdlist[d]->funlist;
  fdlist[d]->funlist=l->next;
  if(fdlist[d]->funlist!=NULL) {
    if(fdlist[d]->funlist->events) {
      set_mask_to_events(d,fdlist[d]->funlist->events);
    }
  }
  else set_mask_to_events(d,fdlist[d]->events);
  free(l);
  return(0);
}

/* close_des() closes and removes descriptor d. */

close_des(d)
     int d;
{
  if(invaliddes(d)) return(1);
  close(d);
  return(remove_des(d));
}

remove_des(d)
     int d;
{
  if(invaliddes(d)) return(1);
  if(debug) fprintf(stderr,"Removing %d\n",d);
  while(fdlist[d]->funlist!=NULL) remove_fun_from_des(d,1);
  if(fdlist[d]->closefun!=NULL) fdlist[d]->closefun(d);
  clear_mask_from_des(d);
  remove_link_from_des(fdlist[d]);
  fdlist[d]=NULL;
  return(0);
}

/* do_select is the function that should be called to do a poll. t
   is an optional timeout value for the select. It returns either 0,
   indicating a timeout, a positive value, indicating that some
   descriptors received events, or a negative value, indicating there
   was an error of some sort (usually caused by closing a descriptor
   without calling close_des() on it; unfortuantely, there's no easy
   way to determine which descriptor is at fault). */

do_select(t)
  struct timeval *t;
{
  fd_set r,w,a;
  int x,i;

  r=read_mask;
  w=write_mask;
  a=except_mask;
  i=select(maxfds,&r,&w,&a,t);
  if(i==0) return(0);
  if(i<0) {
    if(errno!=EINTR) return(-1);
    return(0);
  }
  for(x=minfds;x<maxfds;x++) {
    if(invaliddes(x)) continue;
    if(fdlist[x]->lockcnt) continue;
    if(FD_ISSET(x,&r) || FD_ISSET(x,&w) || FD_ISSET(x,&a) || (fdlist[x]->currevents&8)) {
      if(FD_ISSET(x,&a) && !(fdlist[x]->currevents&4)) {
	syslog(LOG_ERR,"Exception on des %d-closing",x);
	close_des(x);
      }
      else {
	PFI f;

	if(fdlist[x]->funlist!=NULL) f=fdlist[x]->funlist->fun;
	else f=fdlist[x]->defaultfun;
	if(f(x,fdlist[x],0)) close_des(x);
      }
    }
  }
  return(1);
}

struct fdlist *get_fdlist_entry(d)
     int d;
{
  return(fdlist[d]);
}
