#include <stdio.h>
#include <fcntl.h>
#include <sys/file.h>
#include <errno.h>
#include "popd.h"
#include "userinfo.h"
#include "fdlist.h"
#include <syslog.h>

extern int debug;

static struct writebuf *alloc_writebuf();
int write_some_data();
struct fdlist *get_fdlist_entry();
static int alwaysbuffer=1; /* We'll leave this turned on, see whot happens */
static void free_writebuf();
static void append_data_to_writebuf();

/* Set the "watermarks". This causes desc_to_lock to be locked (via
   lock_des) whenever more than HIWATERMARK bytes of data are accumulated
   in the buffers associated with SOCKET. The descriptor is then unlocked
   whenever the data in the buffers goes below LOWATERMARK. */

set_watermark(socket,lowatermark,hiwatermark,desc_to_lock)
     int socket,lowatermark,hiwatermark,desc_to_lock;
{
  struct fdlist *fd;

  fd=get_fdlist_entry(socket);
  if(fd==NULL || fd->userdata==NULL) return(1);
  fd->userdata->lowatermark=lowatermark;
  fd->userdata->hiwatermark=hiwatermark;
  fd->userdata->desc_to_lock=desc_to_lock;
  return(0);
}

/* This removes any watermark info associated with SOCKET. */
clear_watermark(socket)
     int socket;
{
  struct fdlist *fd;

  fd=get_fdlist_entry(socket);
  if(fd==NULL || fd->userdata==NULL) return(1);
  if(fd->userdata->desc_locked) {
    unlockdes(fd->userdata->desc_to_lock);
    fd->userdata->desc_locked=0;
  }
  fd->userdata->desc_to_lock=0-1;
}

/* Do buffered writes. */

write_socket(socket,buf,amt)
     int socket;
     char *buf;
     int amt;
{
  int writeamt;
  struct fdlist *fd;

  fd=get_fdlist_entry(socket);
  if(fd==NULL || fd->userdata==NULL) return(-1);
  /* If there's already buffered writes being done, or if we always
     buffer writes, then append the data onto the write buffer */
  if(fd->userdata->writebuf!=NULL || alwaysbuffer) {
    append_data_to_writebuf(socket,fd,buf,amt);
  }
  else {
    writeamt=write(socket,buf,amt);
    if((writeamt<0)) {
      if(errno!=EWOULDBLOCK) return(-1);
      else writeamt=0;
    }
    else touch_timer(socket);
    if(writeamt!=amt) {
      append_data_to_writebuf(socket,fd,buf+writeamt,amt-writeamt);
    }
  }
  return(amt);
}

/*
  dstart = where the unwritten data starts within the buffer
  amt = the amount of data that hasn't been written in this buffer

  The data doesn't wrap around within the buffer, which is a waste, but
  it simplifies the algorithms a lot.

*/

static void append_data_to_writebuf(socket,fd,buf,amt)
     int socket;
     struct fdlist *fd;
     char *buf;
     int amt;
{
  struct writebuf **p,*ptr;
  int amtfree;

  if(debug>2) printf("Writing %d bytes to socket %d\n",amt,socket);
  p=(&(fd->userdata->writebuf));
  if(*p==NULL) add_fun_to_des(socket,write_some_data,2);
  /* Total amount in the various buffers. */
  fd->userdata->writebufamt+=amt;

  /* If the descriptor isn't already locked, and we're doing watermark
     checking, and we go over the watermark, lock the descriptor. */
  if((!fd->userdata->desc_locked) && (fd->userdata->desc_to_lock>=0) && (fd->userdata->writebufamt>fd->userdata->hiwatermark)) {
    lockdes(fd->userdata->desc_to_lock);
    fd->userdata->desc_locked=1;
  }

  while(amt>0) {
    if((*p)==NULL) {
      if(debug>1) printf("Allocating new buffer\n");
      ptr=(*p=alloc_writebuf());
      ptr->dstart=0;
      ptr->amt=0;
      ptr->next=NULL;
    }
    else {
      ptr=(*p);
      while(ptr->next!=NULL) ptr=ptr->next;
      p=(&(ptr));
    }
    amtfree=sizeof(ptr->buf)-ptr->amt-ptr->dstart;
    if(amtfree>amt) amtfree=amt;
    if(amtfree>0) {
      bcopy(buf,ptr->buf+ptr->dstart+ptr->amt,amtfree);
      buf+=amtfree;
      amt-=amtfree;
      ptr->amt+=amtfree;
      if(debug>1) printf("dstart=%d amt=%d\n",ptr->dstart,ptr->amt);
    }
    p=(&(ptr->next));
  }
}

/* This is one of the thread server functions. It will be called from
   do_select() whenever the socket is ready for writing. It then attempts
   to send data to the socket from its write buffers. It will delete itself
   when it is done writing. */

write_some_data(socket,fd,abrt)
     int socket;
     struct fdlist *fd;
     int abrt;
{
  int writeamt;

  if(abrt) {
    if(fd->userdata->writebuf!=NULL) free_writebufs(&(fd->userdata->writebuf));
    clear_watermark(socket);
    return(0);
  }
  if(fd->userdata->writebuf==NULL) {
    /* All done. */
    remove_fun_from_des(socket,0);
    if(debug>1) printf("writebuf empty when calling, shouldn't happen\n");
    return(0);
  }
  if(fd->userdata->writebuf->amt>0) {
    /* This buffer still has unwritten data in it, so we'll try to write
       it out. The writing algorithm isn't particularly efficient, since
       it's not going to write more than one buffer's worth of data at
       most, and worst case it could write only one byte. */
    writeamt=write(socket,fd->userdata->writebuf->buf+fd->userdata->writebuf->dstart,fd->userdata->writebuf->amt);
    if(writeamt<0) {
      if(errno!=EWOULDBLOCK) {
	syslog(LOG_ERR,"Write error: %m");
	abort_session(fd); /* ??? probably should be removed */
	return(-1);
      }
      return(0);
    }
    else touch_timer(socket); /* Actually wrote some data to the client, so
				 it's active */
    if(writeamt==0) return(0);
    fd->userdata->writebuf->amt -= writeamt;
    fd->userdata->writebuf->dstart+=writeamt;
    fd->userdata->writebufamt-=writeamt;
    if(fd->userdata->desc_locked) {
      /* If we're doing watermarks, and we've gone below the lowatermark,
	 unlock the locked descriptor. */
      if(fd->userdata->writebufamt<=fd->userdata->lowatermark) {
	unlockdes(fd->userdata->desc_to_lock);
	fd->userdata->desc_locked=0;
      }
    }
    if(debug>1) {
      printf("Wrote %d bytes to socket %d, %d left\n",writeamt,socket,fd->userdata->writebuf->amt);
    }
  }
  if(fd->userdata->writebuf->amt<=0) {
    /* This buffer's empty now, we free it up and shift the pointer to the
       next one. */
    struct writebuf *w;

    w=fd->userdata->writebuf;
    fd->userdata->writebuf=w->next;
    if(w->next==NULL) {
      /* No more buffers on the list, so remove the write function from
	 the call list */
      remove_fun_from_des(socket,0);
      if(fd->userdata->closeflag) {
	/* This is set when exiting. */
	if(debug) printf("Closing %d\n",socket);
	close_des(socket);
      }
    }
    free_writebuf(w);
  }
  return(0);
}

static struct writebuf *list=NULL,*staticarea=NULL;
static int numinarea=0;
#define NUMAREASTOALLOCATE 20

static struct writebuf *alloc_writebuf() {
  struct writebuf *res;

  if(list==NULL) {
    if(numinarea<=0) {
      staticarea=tnmalloc(struct writebuf,NUMAREASTOALLOCATE);
      numinarea=NUMAREASTOALLOCATE;
    }
    res=staticarea++;
    numinarea--;
    return(res);
  }
  else {
    res=list;
    list=list->next;
    res->next=NULL;
    return(res);
  }
}

/* Push freed writebuf w onto the list of free write buffers */

static void free_writebuf(w)
     struct writebuf *w;
{
  w->next=list;
  list=w;
}

/* Free all the writebufs from a writebuf list */

free_writebufs(w)
     struct writebuf **w;
{
  struct writebuf *p,*q;

  if(*w==NULL) return;
  else {
    p=(*w);
    (*w)=NULL;
    while(p!=NULL) {
      q=p->next;
      free_writebuf(p);
      p=q;
    }
  }
}
